// ClientConnection.h

#ifndef __ClientConnection_h__
#define __ClientConnection_h__

#include "Vector.h"
#include "EQPacket.h"
#include "EQTxPacket.h"
#include "ThreadSync.h"
#include "EQServerSocketInterface.h"

class CAck
{
	unsigned short	m_ackID;
	CXTime			m_timeReceived;
	bool			m_SYN;

public:
	CAck()
		: m_ackID(0)
		, m_timeReceived(0, 0)
		, m_SYN(false)
	{}

	CAck(unsigned short ackID, bool SYN = false)
		: m_ackID(ackID)
		, m_timeReceived()
		, m_SYN(SYN)
	{}

	CAck(const CAck &model)
		: m_ackID(model.m_ackID)
		, m_timeReceived(model.m_timeReceived)
		, m_SYN(model.m_SYN)
	{}

	operator=(const CAck &o)
	{
		m_ackID = o.m_ackID;
		m_timeReceived = o.m_timeReceived;
	}

	unsigned short getAckID()
	{
		return m_ackID;
	}

	bool isSYN()
	{
		return m_SYN;
	}

	CXTime &getTimeReceived()
	{
		return m_timeReceived;  
	}
};

class CUDPServerSocket;
class CClientConnection : public IClientConnection
{
	friend CUDPServerSocket;

	protected:
		CUDPServerSocket&m_serverSocket;
		sockaddr_in		m_addr;

		CXTime			m_LastSentTick;
		CXTime			m_LastPacketRecieved;
		
		CVector<CEQTxPacket *>	m_SendQueue;
		CVector<CEQPacket *>	m_RecvQueue;
		CVector<CBinary *>		m_OrderedRecvQueue;
		CVector<CAck>			m_AckQueue;

		unsigned short	m_NextRxSeqID;
		unsigned short	m_NextRxFragID;

		unsigned short	m_NextTxFragID;
		unsigned short	m_NextTxSeqID;
		unsigned short	m_NextTxPacketID;
		unsigned char	m_NextTxMsgID;

		bool			m_NeedsSYN;

		CCriticalSection m_lockObj;


	public:
		CClientConnection(CUDPServerSocket &serverSocket, sockaddr_in addr);
		~CClientConnection();

		unsigned long GetIPAddress()
		{
			return m_addr.sin_addr.S_un.S_addr;
		}

		unsigned short GetPort()
		{
			return m_addr.sin_port;
		}

		sockaddr_in getSockAddr()
		{
			return m_addr;
		}

		// returns true if there are packets waiting inthe recieve queue.
		bool IsPacketQueued();

		// retrieves packet from recive queue and removes it too.
		CBinary *GetRecvPacket();

		void Send(CBinary *txMsg, unsigned short headers = 0);

		bool operator==(const CClientConnection &o)
		{
			return this == &o;
		}

	protected:
		void Send(CEQTxPacket *txPacket);

		// called by CUDPServerSocket to give the client a new packet to deal
		// with.
		void newPacket(CEQPacket &pNewPacket);

		// Check to see if any packets weren't ACKed and need resend.
		void DoTick();

		// called when a new message is recieved.
		void MessageRecieved(CBinary *message);

	private:
		void SendImpl(CDatagramPacket *txPacket);

		void ClearTimeOut()
		{
			m_LastPacketRecieved.reset();
		}
};

#endif // __ClientConnection_h__